@@ -28,6 +28,7 @@ enum {
 	GOT_CHUNK,
 	GOT_BODY,
 
+	RESTARTED,
 	FINISHED,
 
 	LAST_SIGNAL
@@ -35,9 +36,11 @@ enum {
 
 static guint signals[LAST_SIGNAL] = { 0 };
 
+static void wrote_body (SoupMessage *req);
 static void got_headers (SoupMessage *req);
 static void got_chunk (SoupMessage *req);
 static void got_body (SoupMessage *req);
+static void restarted (SoupMessage *req);
 static void finished (SoupMessage *req);
 static void free_chunks (SoupMessage *msg);
 
@@ -64,7 +67,7 @@ finalize (GObject *object)
 {
 	SoupMessage *msg = SOUP_MESSAGE (object);
 
-	soup_message_io_cancel (msg);
+	soup_message_io_stop (msg);
 
 	if (msg->priv->uri)
 		soup_uri_free (msg->priv->uri);
@@ -99,9 +102,11 @@ class_init (GObjectClass *object_class)
 	parent_class = g_type_class_ref (PARENT_TYPE);
 
 	/* virtual method definition */
+	message_class->wrote_body   = wrote_body;
 	message_class->got_headers  = got_headers;
 	message_class->got_chunk    = got_chunk;
 	message_class->got_body     = got_body;
+	message_class->restarted    = restarted;
 	message_class->finished     = finished;
 
 	/* virtual method override */
@@ -174,6 +179,14 @@ class_init (GObjectClass *object_class)
 			      soup_marshal_NONE__NONE,
 			      G_TYPE_NONE, 0);
 
+	signals[RESTARTED] =
+		g_signal_new ("restarted",
+			      G_OBJECT_CLASS_TYPE (object_class),
+			      G_SIGNAL_RUN_FIRST,
+			      G_STRUCT_OFFSET (SoupMessageClass, restarted),
+			      NULL, NULL,
+			      soup_marshal_NONE__NONE,
+			      G_TYPE_NONE, 0);
 	signals[FINISHED] =
 		g_signal_new ("finished",
 			      G_OBJECT_CLASS_TYPE (object_class),
@@ -309,6 +322,14 @@ soup_message_wrote_chunk (SoupMessage *msg)
 	g_signal_emit (msg, signals[WROTE_CHUNK], 0);
 }
 
+static void
+wrote_body (SoupMessage *req)
+{
+	g_object_ref (req);
+	soup_message_run_handlers (req, SOUP_HANDLER_POST_REQUEST);
+	g_object_unref (req);
+}
+
 void
 soup_message_wrote_body (SoupMessage *msg)
 {
@@ -369,10 +390,23 @@ soup_message_got_body (SoupMessage *msg)
 	g_signal_emit (msg, signals[GOT_BODY], 0);
 }
 
+static void
+restarted (SoupMessage *req)
+{
+	soup_message_io_stop (req);
+}
+
+void
+soup_message_restarted (SoupMessage *msg)
+{
+	g_signal_emit (msg, signals[RESTARTED], 0);
+}
+
 static void
 finished (SoupMessage *req)
 {
-	soup_message_io_cancel (req);
+	soup_message_io_stop (req);
+	req->status = SOUP_MESSAGE_STATUS_FINISHED;
 }
 
 void
@@ -381,22 +415,6 @@ soup_message_finished (SoupMessage *msg)
 	g_signal_emit (msg, signals[FINISHED], 0);
 }
 
-
-/**
- * soup_message_cancel:
- * @msg: a #SoupMessage currently being processed.
- * 
- * Cancel a running message, and issue completion callback with an
- * status code of %SOUP_STATUS_CANCELLED. If not requeued by the
- * completion callback, the @msg will be destroyed.
- */
-void
-soup_message_cancel (SoupMessage *msg)
-{
-	soup_message_set_status (msg, SOUP_STATUS_CANCELLED);
-	soup_message_finished (msg);
-}
-
 static gboolean
 free_header_list (gpointer name, gpointer vals, gpointer user_data)
 {
@@ -537,7 +555,7 @@ soup_message_cleanup_response (SoupMessage *req)
 
 	soup_message_clear_headers (req->response_headers);
 
-	req->status_code = 0;
+	req->status_code = SOUP_STATUS_NONE;
 	if (req->reason_phrase) {
 		g_free ((char *) req->reason_phrase);
 		req->reason_phrase = NULL;
@@ -561,7 +579,7 @@ soup_message_get_flags (SoupMessage *msg)
 }
 
 void
-soup_message_set_http_version  (SoupMessage *msg, SoupHttpVersion version)
+soup_message_set_http_version (SoupMessage *msg, SoupHttpVersion version)
 {
 	g_return_if_fail (SOUP_IS_MESSAGE (msg));
 
@@ -584,6 +602,10 @@ soup_message_is_keepalive (SoupMessage *msg)
 	c_conn = soup_message_get_header (msg->request_headers, "Connection");
 	s_conn = soup_message_get_header (msg->response_headers, "Connection");
 
+	if (msg->status_code == SOUP_STATUS_OK &&
+	    soup_method_get_id (msg->method) == SOUP_METHOD_ID_CONNECT)
+		return TRUE;
+
 	if (msg->priv->http_version == SOUP_HTTP_1_0) {
 		/* Only persistent if the client requested keepalive
 		 * and the server agreed.
@@ -615,9 +637,9 @@ soup_message_set_uri (SoupMessage *msg, const SoupUri *new_uri)
 
 	if (msg->priv->uri && new_uri) {
 		if (strcmp (msg->priv->uri->host, new_uri->host) != 0)
-			soup_message_io_cancel (msg);
+			soup_message_io_stop (msg);
 	} else if (!new_uri)
-		soup_message_io_cancel (msg);
+		soup_message_io_stop (msg);
 
 	if (msg->priv->uri)
 		soup_uri_free (msg->priv->uri);
